import sys

import logging
from typing import Union
import numpy as np
import jax
import jax.numpy as jnp

from evojax.algo.base import NEAlgorithm
from evojax.util import create_logger


class Sep_CMA_ES(NEAlgorithm):
    """A wrapper around evosax's Sep-CMA-ES.
    Implementation: https://github.com/RobertTLange/evosax/blob/main/evosax/strategies/sep_cma_es.py
    Reference: Ros & Hansen (2008) - https://hal.inria.fr/inria-00287367/document
    """

    def __init__(
        self,
        param_size: int,
        pop_size: int,
        elite_ratio: float = 0.5,
        init_stdev: float = 0.1,
        w_decay: float = 0.0,
        seed: int = 0,
        logger: logging.Logger = None,
    ):
        """Initialization function.

        Args:
            param_size - Parameter size.
            pop_size - Population size.
            elite_ratio - Population elite fraction used for gradient estimate.
            init_stdev - Initial scale of istropic part of covariance.
            w_decay - L2 weight regularization coefficient.
            seed - Random seed for parameters sampling.
            logger - Logger.
        """

        # Delayed importing of evosax

        if sys.version_info.minor < 7:
            print("evosax, which is needed by Sep-CMA-ES, requires python>=3.7")
            print("  please consider upgrading your Python version.")
            sys.exit(1)

        try:
            import evosax
        except ModuleNotFoundError:
            print("You need to install evosax for its Sep-CMA-ES:")
            print("  pip install evosax")
            sys.exit(1)

        # Set up object variables.

        if logger is None:
            self.logger = create_logger(name="Sep_CMA_ES")
        else:
            self.logger = logger

        self.param_size = param_size
        self.pop_size = abs(pop_size)
        self.elite_ratio = elite_ratio
        self.rand_key = jax.random.PRNGKey(seed=seed)

        # Instantiate evosax's ARS strategy
        self.es = evosax.Sep_CMA_ES(
            popsize=pop_size,
            num_dims=param_size,
            elite_ratio=elite_ratio,
        )

        # Set hyperparameters according to provided inputs
        self.es_params = self.es.default_params
        self.es_params["sigma_init"] = init_stdev

        # Initialize the evolution strategy state
        self.rand_key, init_key = jax.random.split(self.rand_key)
        self.es_state = self.es.initialize(init_key, self.es_params)

        # By default evojax assumes maximization of fitness score!
        # Evosax, on the other hand, minimizes!
        self.fit_shaper = evosax.FitnessShaper(w_decay=w_decay, maximize=True)

    def ask(self) -> jnp.ndarray:
        self.rand_key, ask_key = jax.random.split(self.rand_key)
        self.params, self.es_state = self.es.ask(
            ask_key, self.es_state, self.es_params
        )
        return self.params

    def tell(self, fitness: Union[np.ndarray, jnp.ndarray]) -> None:
        # Reshape fitness to conform with evosax minimization
        fit_re = self.fit_shaper.apply(self.params, fitness)
        self.es_state = self.es.tell(
            self.params, fit_re, self.es_state, self.es_params
        )

    @property
    def best_params(self) -> jnp.ndarray:
        return jnp.array(self.es_state["mean"], copy=True)

    @best_params.setter
    def best_params(self, params: Union[np.ndarray, jnp.ndarray]) -> None:
        self.es_state["best_member"] = jnp.array(params, copy=True)
        self.es_state["mean"] = jnp.array(params, copy=True)



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import numpy as np
from typing import Union
from typing import Tuple
from typing import Callable

import jax.numpy as jnp
from jax import tree_util
from flax.core import FrozenDict


def get_params_format_fn(init_params: FrozenDict) -> Tuple[int, Callable]:
    """Return a function that formats the parameters into a correct format."""

    flat, tree = tree_util.tree_flatten(init_params)
    params_sizes = np.cumsum([np.prod(p.shape) for p in flat])

    def params_format_fn(params: jnp.ndarray) -> FrozenDict:
        params = tree_util.tree_multimap(
            lambda x, y: x.reshape(y.shape),
            jnp.split(params, params_sizes, axis=-1)[:-1],
            flat)
        return tree_util.tree_unflatten(tree, params)

    return params_sizes[-1], params_format_fn


def create_logger(name: str,
                  log_dir: str = None,
                  debug: bool = False) -> logging.Logger:
    """Create a logger.

    Args:
        name - Name of the logger.
        log_dir - The logger will also log to an external file in the specified
                  directory if specified.
        debug - If we should log in DEBUG mode.

    Returns:
        logging.RootLogger.
    """

    if log_dir and not os.path.exists(log_dir):
        os.makedirs(log_dir)
    log_format = '%(name)s: %(asctime)s [%(levelname)s] %(message)s'
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.INFO, format=log_format)
    logger = logging.getLogger(name)
    if log_dir:
        log_file = os.path.join(log_dir, '{}.txt'.format(name))
        file_hdl = logging.FileHandler(log_file)
        formatter = logging.Formatter(fmt=log_format)
        file_hdl.setFormatter(formatter)
        logger.addHandler(file_hdl)
    return logger


def load_model(model_dir: str) -> Tuple[np.ndarray, np.ndarray]:
    """Load policy parameters from the specified directory.

    Args:
        model_dir - Directory to load the model from.
    Returns:
        A pair of parameters, the shapes of which are
        (param_size,) and (1 + 2 * obs_params_size,).
    """

    model_file = os.path.join(model_dir, 'model.npz')
    if not os.path.exists(model_file):
        raise ValueError('Model file {} does not exist.')
    with np.load(model_file) as data:
        params = data['params']
        obs_params = data['obs_params']
    return params, obs_params


def save_model(model_dir: str,
               model_name: str,
               params: Union[np.ndarray, jnp.ndarray],
               obs_params: Union[np.ndarray, jnp.ndarray] = None,
               best: bool = False) -> None:
    """Save policy parameters to the specified directory.

    Args:
        model_dir - Directory to save the model.
        model_name - Filename.
        params - The parameters to save.
        obs_params - The observation parameters to save
        best - Whether to save a copy as best.npz.
    """

    model_file = os.path.join(model_dir, '{}.npz'.format(model_name))
    np.savez(model_file,
             params=np.array(params),
             obs_params=np.array(obs_params))
    if best:
        model_file = os.path.join(model_dir, 'best.npz')
        np.savez(model_file,
                 params=np.array(params),
                 obs_params=np.array(obs_params))


def get_tensorboard_log_fn(log_dir: str) -> Callable[[int, jnp.ndarray, str], None]:
    """
    Returns a custom logging function for the `evojax` `Trainer`.
    The function logs rewards after every train/test iteration with Tensorboard.
    It tries to use `tensorflow` or `pytorch` as tensorboard provider.

    Args:
        log_dir - directory to save store the tensorboard logs
    """
    try:
        from torch.utils.tensorboard import SummaryWriter

        def log_with_pytorch(i: int, scores: jnp.ndarray, stage: str):
            with SummaryWriter(log_dir=log_dir) as writer:
                writer.add_scalar(f"{stage}/score_min", scores.min().item(), global_step=i)
                writer.add_scalar(f"{stage}/score_max", scores.max().item(), global_step=i)
                writer.add_scalar(f"{stage}/score_mean", scores.mean().item(), global_step=i)
                writer.add_scalar(f"{stage}/score_std", scores.std().item(), global_step=i)
                writer.add_histogram(f"{stage}/score_distribution", np.array(scores), global_step=i)

        return log_with_pytorch

    except ImportError:
        pass

    try:
        import tensorflow as tf

        def log_with_tf(i: int, scores: jnp.ndarray, stage: str):
            with tf.summary.SummaryWriter(log_dir=log_dir).as_default():
                tf.summary.scalar(f"{stage}/score_min", scores.min().item(), step=i)
                tf.summary.scalar(f"{stage}/score_max", scores.max().item(), step=i)
                tf.summary.scalar(f"{stage}/score_mean", scores.mean().item(), step=i)
                tf.summary.scalar(f"{stage}/score_std", scores.std().item(), step=i)
                tf.summary.histogram(f"{stage}/score_distribution", np.array(scores), step=i)

        return log_with_tf

    except ImportError:
        pass

    raise ImportError("Please install the tensorboard AND (tensorflow OR pytorch) packages to log the rewards to tensorboard")



import logging
import numpy as np
from typing import Union
from typing import Tuple

import jax
import jax.numpy as jnp

from evojax.algo.base import NEAlgorithm
from evojax.util import create_logger


class SimpleGA(NEAlgorithm):
    """A simple genetic algorithm implementing truncation selection."""

    def __init__(self,
                 param_size: int,
                 pop_size: int,
                 truncation_divisor: int = 2,
                 sigma: float = 0.01,
                 seed: int = 0,
                 logger: logging.Logger = None):
        """Initialization function.

        Args:
            param_size - Parameter size.
            pop_size - Population size.
            truncation_divisor - Number by which the population is truncated
                                 every iteration.
            sigma - Variance of normal distribution for parameter perturbation
            seed - Random seed for parameters sampling.
            logger - Logger
        """

        if logger is None:
            self.logger = create_logger(name='SimpleGA')
        else:
            self.logger = logger

        self.param_size = param_size

        self.pop_size = abs(pop_size)
        self.truncation_divisor = abs(truncation_divisor)

        if self.pop_size % 2 == 1:
            self.pop_size += 1
            self._logger.info(
                'Population size should be an even number, set to {}'.format(
                    self.pop_size))

        if self.pop_size % self.truncation_divisor != 0:
            self.truncation_divisor = 2
            self._logger.info(
                'Population size must be a multiple of truncation divisor, \
                 set to {}'.format(self.truncation_divisor))

        self.truncation = self.pop_size // self.truncation_divisor
        self.sigma = sigma

        self.params = jnp.zeros((pop_size, param_size))
        self._best_params = None

        self.rand_key = jax.random.PRNGKey(seed=seed)

        self.jnp_array = jax.jit(jnp.array)

        def ask_fn(key: jnp.ndarray,
                   params: Union[np.ndarray,
                                 jnp.ndarray]) -> Tuple[jnp.ndarray,
                                                        Union[np.ndarray,
                                                              jnp.ndarray]]:

            next_key, sample_key = jax.random.split(key=key, num=2)

            perturbations = jax.random.normal(key=sample_key,
                                              shape=(self.pop_size,
                                                     self.param_size))

            return next_key, params + perturbations * self.sigma

        self.ask_fn = jax.jit(ask_fn)

        def tell_fn(fitness: Union[np.ndarray,
                                   jnp.ndarray],
                    params: Union[np.ndarray,
                                  jnp.ndarray]) -> Union[np.ndarray,
                                                         jnp.ndarray]:

            params = params[fitness.argsort(axis=0)]
            params = params[-self.truncation:].repeat(self.truncation_divisor,
                                                      axis=0)
            return params

        self.tell_fn = jax.jit(tell_fn)

    def ask(self) -> jnp.ndarray:
        self.rand_key, self.params = self.ask_fn(self.rand_key, self.params)
        return self.params

    def tell(self, fitness: Union[np.ndarray, jnp.ndarray]) -> None:
        self.params = self.tell_fn(fitness, self.params)
        self._best_params = self.params[-1]

    @property
    def best_params(self) -> jnp.ndarray:
        return self.jnp_array(self._best_params)

    @best_params.setter
    def best_params(self, params: Union[np.ndarray, jnp.ndarray]) -> None:
        self.params = jnp.repeat(params[None, :], self.pop_size, axis=0)
        self._best_params = jnp.array(params, copy=True)



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the PGPE algorithm in JAX.

Ref: https://github.com/nnaisense/pgpelib/blob/release/pgpelib/pgpe.py
"""

import numpy as np
import logging
from typing import Optional
from typing import Union
from typing import Tuple
from functools import partial

import jax
import jax.numpy as jnp
from jax import random

try:
    from jax.example_libraries import optimizers
except ModuleNotFoundError:
    from jax.experimental import optimizers

from evojax.algo.base import NEAlgorithm
from evojax.util import create_logger


@partial(jax.jit, static_argnums=(1,))
def process_scores(
    x: Union[np.ndarray, jnp.ndarray], use_ranking: bool
) -> jnp.ndarray:
    """Convert fitness scores to rank if necessary."""

    x = jnp.array(x)
    if use_ranking:
        ranks = jnp.zeros(x.size, dtype=int)
        ranks = ranks.at[x.argsort()].set(jnp.arange(x.size)).reshape(x.shape)
        return ranks / ranks.max() - 0.5
    else:
        return x


@jax.jit
def compute_reinforce_update(
    fitness_scores: jnp.ndarray, scaled_noises: jnp.ndarray, stdev: jnp.ndarray
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Compute the updates for the center and the standard deviation."""

    fitness_scores = fitness_scores.reshape((-1, 2))
    baseline = jnp.mean(fitness_scores)
    all_scores = (fitness_scores[:, 0] - fitness_scores[:, 1]).squeeze()
    all_avg_scores = fitness_scores.sum(axis=-1) / 2
    stdev_sq = stdev ** 2.0
    total_mu = scaled_noises * jnp.expand_dims(all_scores, axis=1) * 0.5
    total_sigma = (
        (jnp.expand_dims(all_avg_scores, axis=1) - baseline)
        * (scaled_noises ** 2 - jnp.expand_dims(stdev_sq, axis=0))
        / stdev
    )
    return total_mu.mean(axis=0), total_sigma.mean(axis=0)


@jax.jit
def update_stdev(
    stdev: jnp.ndarray, lr: float, grad: jnp.ndarray, max_change: float
) -> jnp.ndarray:
    """Update (and clip) the standard deviation."""

    allowed_delta = jnp.abs(stdev) * max_change
    min_allowed = stdev - allowed_delta
    max_allowed = stdev + allowed_delta
    return jnp.clip(stdev + lr * grad, min_allowed, max_allowed)


@partial(jax.jit, static_argnums=(3, 4))
def ask_func(
    key: jnp.ndarray,
    stdev: jnp.ndarray,
    center: jnp.ndarray,
    num_directions: int,
    solution_size: int,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """A function that samples a population of parameters from Gaussian."""

    next_key, key = random.split(key)
    scaled_noises = random.normal(key, [num_directions, solution_size]) * stdev
    solutions = jnp.hstack(
        [center + scaled_noises, center - scaled_noises]
    ).reshape(-1, solution_size)
    return next_key, scaled_noises, solutions


class PGPE(NEAlgorithm):
    """Policy Gradient with Parameter-based Exploration (PGPE) algorithm.

    Ref: https://people.idsia.ch/~juergen/icann2008sehnke.pdf
    """

    def __init__(
        self,
        pop_size: int,
        param_size: int,
        init_params: Optional[Union[jnp.ndarray, np.ndarray]] = None,
        optimizer: Optional[str] = None,
        optimizer_config: Optional[dict] = None,
        center_learning_rate: float = 0.15,
        stdev_learning_rate: float = 0.1,
        init_stdev: Union[float, jnp.ndarray, np.ndarray] = 0.1,
        stdev_max_change: float = 0.2,
        solution_ranking: bool = True,
        seed: int = 0,
        logger: logging.Logger = None,
    ):
        """Initialization function.

        Args:
            pop_size - Population size.
            param_size - Parameter size.
            init_params - Initial parameters, all zeros if not given.
            optimizer - Possible values are {None, 'adam', 'clipup'}.
            optimizer_config - Configurations specific to the optimizer.
                               For None: No configuration is required.
                               For Adam: {'epsilon', 'beta1', 'beta2'}.
                               For ClipUp: {'momentum', 'max_speed'}.
            center_learning_rate - Learning rate for the Gaussian mean.
            stdev_learning_rate - Learning rate for the Gaussian stdev.
            init_stdev - Initial stdev for the Gaussian distribution.
            stdev_max_change - Maximum allowed change for stdev in abs values.
            solution_ranking - Should we treat the fitness as rankings or not.
            seed - Random seed for parameters sampling.
        """

        if logger is None:
            self._logger = create_logger(name="PGPE")
        else:
            self._logger = logger

        self.pop_size = abs(pop_size)
        if self.pop_size % 2 == 1:
            self.pop_size += 1
            self._logger.info(
                "Population size should be an even number, set to {}".format(
                    self.pop_size
                )
            )
        self._num_directions = self.pop_size // 2

        if init_params is None:
            self._center = np.zeros(abs(param_size))
        else:
            self._center = init_params
        self._center = jnp.array(self._center)
        if isinstance(init_stdev, float):
            self._stdev = np.ones(abs(param_size)) * abs(init_stdev)
        self._stdev = jnp.array(self._stdev)

        self._center_lr = abs(center_learning_rate)
        self._stdev_lr = abs(stdev_learning_rate)
        self._stdev_max_change = abs(stdev_max_change)
        self._solution_ranking = solution_ranking

        if optimizer_config is None:
            optimizer_config = {}
        decay_coef = optimizer_config.get("center_lr_decay_coef", 1.0)
        self._lr_decay_steps = optimizer_config.get(
            "center_lr_decay_steps", 1000
        )

        if optimizer == "adam":
            opt_init, opt_update, get_params = optimizers.adam(
                step_size=lambda x: self._center_lr * jnp.power(decay_coef, x),
                b1=optimizer_config.get("beta1", 0.9),
                b2=optimizer_config.get("beta2", 0.999),
                eps=optimizer_config.get("epsilon", 1e-8),
            )
        elif optimizer == "clipup":
            opt_init, opt_update, get_params = clipup(
                step_size=lambda x: self._center_lr * jnp.power(decay_coef, x),
                momentum=optimizer_config.get("momentum", 0.9),
                max_speed=optimizer_config.get("max_speed", 0.15),
                fix_gradient_size=optimizer_config.get(
                    "fix_gradient_size", True
                ),
            )
        else:
            opt_init, opt_update, get_params = optimizers.sgd(
                step_size=lambda x: self._center_lr * jnp.power(decay_coef, x),
            )
        self._t = 0
        self._opt_state = jax.jit(opt_init)(self._center)
        self._opt_update = jax.jit(opt_update)
        self._get_params = jax.jit(get_params)

        self._key = random.PRNGKey(seed=seed)
        self._solutions = None
        self._scaled_noises = None

    def ask(self) -> jnp.ndarray:
        self._key, self._scaled_noises, self._solutions = ask_func(
            self._key,
            self._stdev,
            self._center,
            self._num_directions,
            self._center.size,
        )
        return self._solutions

    def tell(self, fitness: Union[np.ndarray, jnp.ndarray]) -> None:
        fitness_scores = process_scores(fitness, self._solution_ranking)
        grad_center, grad_stdev = compute_reinforce_update(
            fitness_scores=fitness_scores,
            scaled_noises=self._scaled_noises,
            stdev=self._stdev,
        )
        self._opt_state = self._opt_update(
            self._t // self._lr_decay_steps, -grad_center, self._opt_state
        )
        self._t += 1
        self._center = self._get_params(self._opt_state)
        self._stdev = update_stdev(
            stdev=self._stdev,
            lr=self._stdev_lr,
            max_change=self._stdev_max_change,
            grad=grad_stdev,
        )

    @property
    def best_params(self) -> jnp.ndarray:
        return jnp.array(self._center, copy=True)

    @best_params.setter
    def best_params(self, params: Union[np.ndarray, jnp.ndarray]) -> None:
        self._center = jnp.array(params, copy=True)


@optimizers.optimizer
def clipup(
    step_size: float,
    momentum: float = 0.9,
    max_speed: float = 0.15,
    fix_gradient_size: bool = True,
):
    """Construct optimizer triple for ClipUp."""

    step_size = optimizers.make_schedule(step_size)

    def init(x0):
        v0 = jnp.zeros_like(x0)
        return x0, v0

    def update(i, g, state):
        x, v = state
        g = jax.lax.cond(
            fix_gradient_size,
            lambda p: p / jnp.sqrt(jnp.sum(p * p)),
            lambda p: p,
            g,
        )
        step = g * step_size(i)
        v = momentum * v + step
        # Clip.
        length = jnp.sqrt(jnp.sum(v * v))
        v = jax.lax.cond(
            length > max_speed, lambda p: p * max_speed / length, lambda p: p, v
        )
        return x - v, v

    def get_params(state):
        x, _ = state
        return x

    return init, update, get_params



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A seq2seq task (simple addition).

Ref: https://github.com/google/flax/tree/main/examples/seq2seq
"""

import numpy as np
from typing import Tuple

import jax
import jax.numpy as jnp
from jax import random
from flax.struct import dataclass

from evojax.task.base import VectorizedTask
from evojax.task.base import TaskState


@dataclass
class State(TaskState):
    obs: jnp.ndarray
    labels: jnp.ndarray


class CharacterTable(object):
    """Encode/decodes between strings and integer representations."""

    def __init__(self):
        self._chars = '0123456789+= '
        self.pad_id = len(self._chars)
        self.eos_id = self.pad_id + 1
        self.vocab_size = len(self._chars) + 2
        self._indices_char = dict(
            (idx, ch) for idx, ch in enumerate(self._chars))
        self._indices_char[self.pad_id] = '_'

    def encode(self, inputs: jnp.ndarray) -> jnp.ndarray:
        return jnp.concatenate([inputs, jnp.array([self.eos_id])])

    def decode(self, inputs):
        """Decode from list of integers to string."""
        chars = []
        for elem in inputs.tolist():
            if elem == self.eos_id:
                break
            chars.append(self._indices_char[elem])
        return ''.join(chars)


class Seq2seqTask(VectorizedTask):
    """Seq2seq task: encoder's input is "x+y", decoder's output is "=z"."""

    def __init__(self,
                 batch_size: int = 128,
                 max_len_query_digit: int = 3,
                 test: bool = False):

        char_table = CharacterTable()
        max_input_len = max_len_query_digit + 2 + 2
        max_output_len = max_len_query_digit + 3
        max_num = pow(10, max_len_query_digit)
        self.obs_shape = tuple([max_input_len, char_table.vocab_size])
        self.act_shape = tuple([max_output_len, char_table.vocab_size])
        self.max_steps = 1

        def encode_onehot(batch_inputs, max_len):
            def encode_str(s):
                tokens = char_table.encode(s)
                org_len = len(tokens)
                tokens = jnp.pad(
                    tokens, [(0, max_len - org_len)], mode='constant')
                return jax.nn.one_hot(
                    tokens, char_table.vocab_size, dtype=jnp.float32)
            return jnp.array([encode_str(inp) for inp in batch_inputs])

        def decode_onehot(batch_inputs):
            return np.array(list(map(
                lambda x: char_table.decode(x.argmax(axis=-1)), batch_inputs)))
        self.decode_embeddings = decode_onehot

        def breakdown_int(n):
            return jnp.where(
                n == 0,
                jnp.zeros(max_len_query_digit),
                jnp.array([(n // (10**i)) % 10
                           for i in range(max_len_query_digit - 1, -1, -1)]))

        def get_batch_data(key):
            keys = random.split(key, 2)
            add_op1 = random.randint(
                keys[0], minval=0, maxval=100, shape=(batch_size,))
            add_op2 = random.randint(
                keys[1], minval=0, maxval=max_num, shape=(batch_size,))
            for op1, op2 in zip(add_op1, add_op2):
                inputs = jnp.concatenate([
                    breakdown_int(op1)[1:],
                    jnp.array([10, ]),   # 10 is '+'
                    breakdown_int(op2)], axis=0)
                outputs = jnp.concatenate([
                    jnp.array([11, ]),  # 11 is '='
                    breakdown_int(op1 + op2)], axis=0)
                yield inputs, outputs

        def reset_fn(key):
            inputs, outputs = zip(*get_batch_data(key[0]))
            batch_data = encode_onehot(inputs, max_input_len)
            batch_labels = encode_onehot(outputs, max_output_len)
            return State(
                obs=jnp.repeat(batch_data[None, :], key.shape[0], axis=0),
                labels=jnp.repeat(batch_labels[None, :], key.shape[0], axis=0))
        self._reset_fn = jax.jit(reset_fn)

        def get_sequence_lengths(sequence_batch):
            # sequence_batch.shape = (batch_size, seq_length, vocab_size)
            eos_row = sequence_batch[:, :, char_table.eos_id]
            eos_idx = jnp.argmax(eos_row, axis=-1)
            return jnp.where(
                eos_row[jnp.arange(eos_row.shape[0]), eos_idx],
                eos_idx + 1, sequence_batch.shape[1])

        def mask_sequences(sequence_batch, lengths):
            return sequence_batch * (
                lengths[:, np.newaxis] >
                np.arange(sequence_batch.shape[1])[np.newaxis])

        def cross_entropy_loss(logits, labels, lengths):
            xe = jnp.sum(jax.nn.log_softmax(logits) * labels, axis=-1)
            return -jnp.mean(mask_sequences(xe, lengths))

        def step_fn(state, action):
            labels = state.labels[:, 1:]
            lengths = get_sequence_lengths(labels)
            if test:
                token_acc = jnp.argmax(action, -1) == jnp.argmax(labels, -1)
                sequence_acc = (jnp.sum(
                    mask_sequences(token_acc, lengths), axis=-1) == lengths)
                reward = jnp.mean(sequence_acc)
            else:
                reward = -cross_entropy_loss(action, labels, lengths)
            done = jnp.ones((), dtype=jnp.int32)
            return state, reward, done
        self._step_fn = jax.jit(jax.vmap(step_fn))

    def reset(self, key: jnp.array) -> State:
        return self._reset_fn(key)

    def step(self,
             state: State,
             action: jnp.ndarray) -> Tuple[State, jnp.ndarray, jnp.ndarray]:
        return self._step_fn(state, action)



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This is a wrapper of the CMA-ES algorithm.

This is for users who want to use CMA-ES before we release a pure JAX version.

CMA-ES paper: https://arxiv.org/abs/1604.00772
"""

import logging
import cma
import numpy as np
from typing import Union

import jax
import jax.numpy as jnp

from evojax.algo.base import NEAlgorithm
from evojax.util import create_logger


class CMA(NEAlgorithm):
    """A wrapper of CMA-ES."""

    def __init__(self,
                 param_size: int,
                 pop_size: int,
                 init_stdev: float = 0.1,
                 seed: int = 0,
                 logger: logging.Logger = None):
        if logger is None:
            self.logger = create_logger(name='CMA')
        else:
            self.logger = logger
        self.pop_size = pop_size
        self.cma = cma.CMAEvolutionStrategy(
            x0=np.zeros(param_size),
            sigma0=init_stdev,
            inopts={
                'popsize': pop_size,
                'seed': seed if seed > 0 else 42,
                'randn': np.random.randn,
            },
        )
        self.params = None
        self._best_params = None

        self.jnp_array = jax.jit(jnp.array)
        self.jnp_stack = jax.jit(jnp.stack)

    def ask(self) -> jnp.ndarray:
        self.params = self.cma.ask()
        return self.jnp_stack(self.params)

    def tell(self, fitness: Union[np.ndarray, jnp.ndarray]) -> None:
        self.cma.tell(self.params, -np.array(fitness))
        self._best_params = np.array(self.cma.result.xfavorite)

    @property
    def best_params(self) -> jnp.ndarray:
        return self.jnp_array(self._best_params)

    @best_params.setter
    def best_params(self, params: Union[np.ndarray, jnp.ndarray]) -> None:
        self._best_params = np.array(params)
        self.cma.x0 = self._best_params.copy()



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import numpy as np
from typing import Tuple

import jax
import jax.numpy as jnp
from flax.struct import dataclass

from evojax.task.base import TaskState
from evojax.task.base import VectorizedTask


@dataclass
class State(TaskState):
    obs: jnp.ndarray


class MDKP(VectorizedTask):
    """Multi-Dimensional Knapsack Problem (MDKP).

    Many real-world problems are MDKP in different contexts, and this task
    allows a user to simulate one by using their own data in the form of a csv
    file. If a csv file is not specified, we use synthesized data.

    The format of the csv file is:
        1st line contains the total cap of each attribute (M columns):
            attr1_cap, attr2_cap, ..., attrM_cap
        2nd line and beyond have details for each item (M+1 columns, N rows):
            item_attr1, item1_attr2, ..., item1_attrM, item1_value
            ...
            itemN_attr1, itemN_attr2, ..., itemN_attrM, itemN_value

    A valid solution should select k items that maximizes sum(item_k_value),
    while keeping sum(item_attr_i) < attr_i_cap for all i in [1, M].
    """

    def __init__(self, csv_file=None, test=False):
        self.max_steps = 1
        self.test = test

        if csv_file is None or not os.path.exists(csv_file):
            rnd = np.random.RandomState(seed=0)
            num_attrs = 3
            num_items = 3000
            upper_bound = 100
            data = rnd.uniform(1, upper_bound, size=(num_items, num_attrs + 1))
            caps = rnd.uniform(
                num_items / 3, num_items / 2, num_attrs) * upper_bound
        else:
            try:
                import pandas as pd
            except ModuleNotFoundError:
                print('This task requires pandas, '
                      'run "pip install -U pandas" to install.')
                sys.exit(1)

            data = pd.read_csv(csv_file, header=None).values
            num_items, num_attrs = data.shape
            num_items -= 1  # Exclude the 1st line.
            num_attrs -= 1  # Exclude the value column.
            caps = data[0][:num_attrs]
            data = data[1:]

        self.num_items = num_items
        self.num_attrs = num_attrs
        self.data = data
        self.caps = caps
        data = jnp.array(data)
        caps = jnp.array(caps)
        self.obs_shape = tuple([num_items, num_attrs + 1])
        self.act_shape = tuple([num_items, ])

        def reset_fn(key):
            return State(obs=data)
        self._reset_fn = jax.jit(jax.vmap(reset_fn))

        def step_fn(state, action):
            selection = jnp.where(action > 0.5, 1, 0)
            total_attr_and_values = (state.obs * selection[:, None]).sum(axis=0)
            reward = total_attr_and_values[-1]
            # If there is any attribute violation, we set reward to zero.
            violation = jnp.prod(jnp.where(
                total_attr_and_values[:num_attrs] > caps, 0, 1))
            reward = reward * violation
            return state, reward, jnp.ones(())
        self._step_fn = jax.jit(jax.vmap(step_fn))

    def reset(self, key: jnp.array) -> TaskState:
        return self._reset_fn(key)

    def step(self,
             state: TaskState,
             action: jnp.ndarray) -> Tuple[TaskState, jnp.ndarray, jnp.ndarray]:
        return self._step_fn(state, action)



import sys

import logging
from typing import Union
import numpy as np
import jax
import jax.numpy as jnp

from evojax.algo.base import NEAlgorithm
from evojax.util import create_logger


class ARS(NEAlgorithm):
    """A wrapper around evosax's Augmented Random Search.
    Implementation: https://github.com/RobertTLange/evosax/blob/main/evosax/strategies/ars.py
    Reference: Mania et al. (2018) - https://arxiv.org/pdf/1803.07055.pdf

    NOTE: More details on the optimizer configuration can be found here
    https://github.com/RobertTLange/evosax/blob/main/evosax/utils/optimizer.py
    """

    def __init__(
        self,
        param_size: int,
        pop_size: int,
        elite_ratio: float = 0.2,
        optimizer: str = "clipup",
        optimizer_config: dict = {
            "lrate_init": 0.15,  # Initial learning rate
            "lrate_decay": 0.999,  # Multiplicative decay factor
            "lrate_limit": 0.05,  # Smallest possible lrate
            "max_speed": 0.3,  # Max. clipping velocity
            "momentum": 0.9,  # Momentum coefficient
        },
        init_stdev: float = 0.01,
        decay_stdev: float = 0.999,
        limit_stdev: float = 0.001,
        w_decay: float = 0.0,
        seed: int = 0,
        logger: logging.Logger = None,
    ):
        """Initialization function.

        Args:
            param_size - Parameter size.
            pop_size - Population size.
            elite_ratio - Population elite fraction used for gradient estimate.
            optimizer - Optimizer name ("sgd", "adam", "rmsprop", "clipup").
            optimizer_config - Configuration of optimizer hyperparameters.
            init_stdev - Initial scale of Gaussian perturbation.
            decay_stdev - Multiplicative scale decay between tell iterations.
            limit_stdev - Smallest scale (clipping limit).
            w_decay - L2 weight regularization coefficient.
            seed - Random seed for parameters sampling.
            logger - Logger.
        """

        # Delayed importing of evosax

        if sys.version_info.minor < 7:
            print(
                "evosax, which is needed by Augmented Random Search, requires"
                " python>=3.7"
            )
            print("  please consider upgrading your Python version.")
            sys.exit(1)

        try:
            import evosax
        except ModuleNotFoundError:
            print("You need to install evosax for its Augmented Random Search:")
            print("  pip install evosax")
            sys.exit(1)

        # Set up object variables.

        if logger is None:
            self.logger = create_logger(name="ARS")
        else:
            self.logger = logger

        self.param_size = param_size
        self.pop_size = abs(pop_size)
        self.elite_ratio = elite_ratio
        self.rand_key = jax.random.PRNGKey(seed=seed)

        # Instantiate evosax's ARS strategy
        self.es = evosax.ARS(
            popsize=pop_size,
            num_dims=param_size,
            elite_ratio=elite_ratio,
            opt_name=optimizer,
        )

        # Set hyperparameters according to provided inputs
        self.es_params = self.es.default_params
        for k, v in optimizer_config.items():
            self.es_params[k] = v
        self.es_params["sigma_init"] = init_stdev
        self.es_params["sigma_decay"] = decay_stdev
        self.es_params["sigma_limit"] = limit_stdev
        self.es_params["init_min"] = 0.0
        self.es_params["init_max"] = 0.0

        # Initialize the evolution strategy state
        self.rand_key, init_key = jax.random.split(self.rand_key)
        self.es_state = self.es.initialize(init_key, self.es_params)

        # By default evojax assumes maximization of fitness score!
        # Evosax, on the other hand, minimizes!
        self.fit_shaper = evosax.FitnessShaper(w_decay=w_decay, maximize=True)

    def ask(self) -> jnp.ndarray:
        self.rand_key, ask_key = jax.random.split(self.rand_key)
        self.params, self.es_state = self.es.ask(
            ask_key, self.es_state, self.es_params
        )
        return self.params

    def tell(self, fitness: Union[np.ndarray, jnp.ndarray]) -> None:
        # Reshape fitness to conform with evosax minimization
        fit_re = self.fit_shaper.apply(self.params, fitness)
        self.es_state = self.es.tell(
            self.params, fit_re, self.es_state, self.es_params
        )

    @property
    def best_params(self) -> jnp.ndarray:
        return jnp.array(self.es_state["mean"], copy=True)

    @best_params.setter
    def best_params(self, params: Union[np.ndarray, jnp.ndarray]) -> None:
        self.es_state["best_member"] = jnp.array(params, copy=True)
        self.es_state["mean"] = jnp.array(params, copy=True)



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""AttentionNeuron augmented MLP.

Ref: https://attentionneuron.github.io/
"""

import logging
from typing import Tuple
import numpy as np

import jax
import jax.numpy as jnp
from jax import random
from flax import linen as nn
from flax.struct import dataclass

from evojax.task.base import TaskState
from evojax.policy.base import PolicyNetwork
from evojax.policy.base import PolicyState
from evojax.util import create_logger
from evojax.util import get_params_format_fn


@dataclass
class State(PolicyState):
    prev_actions: jnp.ndarray
    lstm_states: Tuple[jnp.ndarray, jnp.ndarray]


def pos_table(n, dim):
    """Create a table of positional encodings."""

    def get_angle(x, h):
        return x / np.power(10000, 2 * (h // 2) / dim)

    def get_angle_vec(x):
        return [get_angle(x, j) for j in range(dim)]

    tab = np.array([get_angle_vec(i) for i in range(n)]).astype(float)
    tab[:, 0::2] = np.sin(tab[:, 0::2])
    tab[:, 1::2] = np.cos(tab[:, 1::2])
    return tab


class AttentionNeuronMLP(nn.Module):

    act_dim: int
    msg_dim: int
    hidden_dim: int
    pos_em_dim: int
    pos_em: np.ndarray

    @nn.compact
    def __call__(self, obs, prev_act, lstm_h):
        obs_dim = obs.shape[0]
        # obs.shape: (obs_dim,) to (obs_dim, 1)
        obs = jnp.expand_dims(obs, axis=-1)
        # prev_act.shape: (act_dim,) to (obs_dim, act_dim)
        prev_act = jnp.repeat(
            jnp.expand_dims(prev_act, axis=0), repeats=obs_dim, axis=0)
        # x_aug.shape: (obs_dim, act_dim + 1)
        x_aug = jnp.concatenate([obs, prev_act], axis=-1)

        # q.shape: (hidden_dim, msg_dim)
        q = nn.Dense(self.msg_dim)(self.pos_em)

        # x_key.shape: (obs_dim, pos_em_dim)
        new_lstm_h, x_key = nn.LSTMCell()(lstm_h, x_aug)
        # k.shape: (obs_dim, msg_dim)
        k = nn.Dense(self.msg_dim)(x_key)

        # att.shape: (hidden_dim, obs_dim)
        att = nn.tanh(jnp.matmul(q, k.T))
        # x.shape: (hidden_dim,)
        x = nn.tanh(jnp.matmul(att, obs).squeeze(-1))
        # act.shape: (act_dim,)
        act = nn.tanh(nn.Dense(self.act_dim)(x))

        return act, new_lstm_h


class PermutationInvariantPolicy(PolicyNetwork):
    """A permutation invariant model."""

    def __init__(self,
                 act_dim: int,
                 hidden_dim: int,
                 msg_dim: int = 32,
                 pos_em_dim: int = 8,
                 logger: logging.Logger = None):
        if logger is None:
            self._logger = create_logger(name='PermutationInvariantPolicy')
        else:
            self._logger = logger

        self.act_dim = act_dim
        self.pos_em_dim = pos_em_dim
        model = AttentionNeuronMLP(
            act_dim=act_dim,
            msg_dim=msg_dim,
            pos_em_dim=pos_em_dim,
            hidden_dim=hidden_dim,
            pos_em=pos_table(hidden_dim, pos_em_dim),
        )
        obs_dim = 5
        params = model.init(
            random.PRNGKey(0),
            obs=jnp.ones([obs_dim, ]),
            prev_act=jnp.zeros([act_dim, ]),
            lstm_h=(jnp.zeros([obs_dim, pos_em_dim]),
                    jnp.zeros([obs_dim, pos_em_dim])),
        )
        self.num_params, format_params_fn = get_params_format_fn(params)
        self._logger.info(
            'PermutationInvariantPolicy.num_params = {}'.format(self.num_params)
        )
        self._format_params_fn = jax.vmap(format_params_fn)
        self._forward_fn = jax.vmap(model.apply)

    def reset(self, states: TaskState) -> PolicyState:
        keys = random.split(random.PRNGKey(0), states.obs.shape[0])
        b_size, obs_dim = states.obs.shape
        prev_act = jnp.zeros([b_size, self.act_dim])
        lstm_h = (jnp.zeros([b_size, obs_dim, self.pos_em_dim]),
                  jnp.zeros([b_size, obs_dim, self.pos_em_dim]))
        return State(keys=keys, prev_actions=prev_act, lstm_states=lstm_h)

    def get_actions(self,
                    t_tasks: TaskState,
                    params: jnp.ndarray,
                    p_states: State) -> Tuple[jnp.ndarray, State]:
        params = self._format_params_fn(params)
        act, lstm_h = self._forward_fn(
            params,
            obs=t_tasks.obs,
            prev_act=p_states.prev_actions,
            lstm_h=p_states.lstm_states)
        return act, State(p_states.keys, act, lstm_h)



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import NEAlgorithm
from .cma_wrapper import CMA
from .pgpe import PGPE
from .ars import ARS
from .simple_ga import SimpleGA
from .open_es import OpenES
from .cma_evosax import CMA_ES
from .sep_cma_es import Sep_CMA_ES

Strategies = {
    "CMA": CMA,
    "PGPE": PGPE,
    "SimpleGA": SimpleGA,
    "ARS": ARS,
    "OpenES": OpenES,
    "CMA_ES": CMA_ES,
    "Sep_CMA_ES": Sep_CMA_ES,
}

__all__ = [
    "NEAlgorithm",
    "CMA",
    "PGPE",
    "ARS",
    "SimpleGA",
    "OpenES",
    "CMA_ES",
    "Sep_CMA_ES",
    "Strategies",
]



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import numpy as np
from typing import Tuple

import jax
import jax.numpy as jnp
from jax import random
from flax.struct import dataclass

from evojax.task.base import VectorizedTask
from evojax.task.base import TaskState


@dataclass
class State(TaskState):
    obs: jnp.ndarray
    labels: jnp.ndarray


def sample_batch(key: jnp.ndarray,
                 data: jnp.ndarray,
                 labels: jnp.ndarray,
                 batch_size: int) -> Tuple:
    ix = random.choice(
        key=key, a=data.shape[0], shape=(batch_size,), replace=False)
    return (jnp.take(data, indices=ix, axis=0),
            jnp.take(labels, indices=ix, axis=0))


def loss(prediction: jnp.ndarray, target: jnp.ndarray) -> jnp.float32:
    target = jax.nn.one_hot(target, 10)
    return -jnp.mean(jnp.sum(prediction * target, axis=1))


def accuracy(prediction: jnp.ndarray, target: jnp.ndarray) -> jnp.float32:
    predicted_class = jnp.argmax(prediction, axis=1)
    return jnp.mean(predicted_class == target)


class MNIST(VectorizedTask):
    """MNIST classification task."""

    def __init__(self,
                 batch_size: int = 1024,
                 test: bool = False):

        self.max_steps = 1
        self.obs_shape = tuple([28, 28, 1])
        self.act_shape = tuple([10, ])

        # Delayed importing of torchvision

        try:
            from torchvision import datasets
        except ModuleNotFoundError:
            print('You need to install torchvision for this task.')
            print('  pip install torchvision')
            sys.exit(1)

        dataset = datasets.MNIST('./data', train=not test, download=True)
        data = np.expand_dims(dataset.data.numpy() / 255., axis=-1)
        labels = dataset.targets.numpy()

        def reset_fn(key):
            if test:
                batch_data, batch_labels = data, labels
            else:
                batch_data, batch_labels = sample_batch(
                    key, data, labels, batch_size)
            return State(obs=batch_data, labels=batch_labels)
        self._reset_fn = jax.jit(jax.vmap(reset_fn))

        def step_fn(state, action):
            if test:
                reward = accuracy(action, state.labels)
            else:
                reward = -loss(action, state.labels)
            return state, reward, jnp.ones(())
        self._step_fn = jax.jit(jax.vmap(step_fn))

    def reset(self, key: jnp.ndarray) -> State:
        return self._reset_fn(key)

    def step(self,
             state: TaskState,
             action: jnp.ndarray) -> Tuple[TaskState, jnp.ndarray, jnp.ndarray]:
        return self._step_fn(state, action)



# Copyright 2022 The EvoJAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from typing import Optional, Callable

import jax.numpy as jnp
import numpy as np

from evojax.task.base import VectorizedTask
from evojax.policy import PolicyNetwork
from evojax.algo import NEAlgorithm
from evojax.sim_mgr import SimManager
from evojax.obs_norm import ObsNormalizer
from evojax.util import create_logger
from evojax.util import load_model
from evojax.util import save_model


class Trainer(object):
    """A trainer that organizes the training logistics."""

    def __init__(self,
                 policy: PolicyNetwork,
                 solver: NEAlgorithm,
                 train_task: VectorizedTask,
                 test_task: VectorizedTask,
                 max_iter: int = 1000,
                 log_interval: int = 20,
                 test_interval: int = 100,
                 n_repeats: int = 1,
                 test_n_repeats: int = 1,
                 n_evaluations: int = 100,
                 seed: int = 42,
                 debug: bool = False,
                 normalize_obs: bool = False,
                 model_dir: str = None,
                 log_dir: str = None,
                 logger: logging.Logger = None,
                 log_scores_fn: Optional[Callable[[int, jnp.ndarray, str], None]] = None):
        """Initialization.

        Args:
            policy - The policy network to use.
            solver - The ES algorithm for optimization.
            train_task - The task for training.
            test_task - The task for evaluation.
            max_iter - Maximum number of training iterations.
            log_interval - Interval for logging.
            test_interval - Interval for tests.
            n_repeats - Number of rollout repetitions.
            n_evaluations - Number of tests to conduct.
            seed - Random seed to use.
            debug - Whether to turn on the debug flag.
            normalize_obs - Whether to use an observation normalizer.
            model_dir - Directory to save/load model.
            log_dir - Directory to dump logs.
            logger - Logger.
            log_scores_fn - custom function to log the scores array. Expects input:
                `current_iter`: int, `scores`: jnp.ndarray, 'stage': str = "train" | "test"
        """

        if logger is None:
            self._logger = create_logger(
                name='Trainer', log_dir=log_dir, debug=debug)
        else:
            self._logger = logger

        self._log_interval = log_interval
        self._test_interval = test_interval
        self._max_iter = max_iter
        self.model_dir = model_dir
        self._log_dir = log_dir

        self._log_scores_fn = log_scores_fn or (lambda x, y, z: None)

        self._obs_normalizer = ObsNormalizer(
            obs_shape=train_task.obs_shape,
            dummy=not normalize_obs,
        )

        self.solver = solver
        self.sim_mgr = SimManager(
            n_repeats=n_repeats,
            test_n_repeats=test_n_repeats,
            pop_size=solver.pop_size,
            n_evaluations=n_evaluations,
            policy_net=policy,
            train_vec_task=train_task,
            valid_vec_task=test_task,
            seed=seed,
            obs_normalizer=self._obs_normalizer,
            logger=self._logger,
        )

    def run(self, demo_mode: bool = False) -> float:
        """Start the training / test process."""

        if self.model_dir is not None:
            params, obs_params = load_model(model_dir=self.model_dir)
            self.sim_mgr.obs_params = obs_params
            self._logger.info(
                'Loaded model parameters from {}.'.format(self.model_dir))
        else:
            params = None

        if demo_mode:
            if params is None:
                raise ValueError('No policy parameters to evaluate.')
            self._logger.info('Start to test the parameters.')
            scores = np.array(
                self.sim_mgr.eval_params(params=params, test=True))
            self._logger.info(
                '[TEST] #tests={0}, max={1:.4f}, avg={2:.4f}, min={3:.4f}, '
                'std={4:.4f}'.format(scores.size, scores.max(), scores.mean(),
                                     scores.min(), scores.std()))
            return scores.mean()
        else:
            self._logger.info(
                'Start to train for {} iterations.'.format(self._max_iter))

            if params is not None:
                # Continue training from the breakpoint.
                self.solver.best_params = params

            best_score = -float('Inf')

            for i in range(self._max_iter):
                start_time = time.perf_counter()
                params = self.solver.ask()
                self._logger.debug('solver.ask time: {0:.4f}s'.format(
                    time.perf_counter() - start_time))

                start_time = time.perf_counter()
                scores = self.sim_mgr.eval_params(params=params, test=False)
                self._logger.debug('sim_mgr.eval_params time: {0:.4f}s'.format(
                    time.perf_counter() - start_time))

                start_time = time.perf_counter()
                self.solver.tell(fitness=scores)
                self._logger.debug('solver.tell time: {0:.4f}s'.format(
                    time.perf_counter() - start_time))

                if i > 0 and i % self._log_interval == 0:
                    scores = np.array(scores)
                    self._logger.info(
                        'Iter={0}, size={1}, max={2:.4f}, '
                        'avg={3:.4f}, min={4:.4f}, std={5:.4f}'.format(
                            i, scores.size, scores.max(), scores.mean(),
                            scores.min(), scores.std()))
                    self._log_scores_fn(i, scores, "train")

                if i > 0 and i % self._test_interval == 0:
                    best_params = self.solver.best_params
                    test_scores = self.sim_mgr.eval_params(
                        params=best_params, test=True)
                    self._logger.info(
                        '[TEST] Iter={0}, #tests={1}, max={2:.4f} avg={3:.4f}, '
                        'min={4:.4f}, std={5:.4f}'.format(
                            i, test_scores.size, test_scores.max(),
                            test_scores.mean(), test_scores.min(),
                            test_scores.std()))
                    self._log_scores_fn(i, test_scores, "test")
                    mean_test_score = test_scores.mean()
                    save_model(
                        model_dir=self._log_dir,
                        model_name='iter_{}'.format(i),
                        params=best_params,
                        obs_params=self.sim_mgr.obs_params,
                        best=mean_test_score > best_score,
                    )
                    best_score = max(best_score, mean_test_score)

            # Test and save the final model.
            best_params = self.solver.best_params
            test_scores = self.sim_mgr.eval_params(
                params=best_params, test=True)
            self._logger.info(
                '[TEST] Iter={0}, #tests={1}, max={2:.4f}, avg={3:.4f}, '
                'min={4:.4f}, std={5:.4f}'.format(
                    self._max_iter, test_scores.size, test_scores.max(),
                    test_scores.mean(), test_scores.min(), test_scores.std()))
            mean_test_score = test_scores.mean()
            save_model(
                model_dir=self._log_dir,
                model_name='final',
                params=best_params,
                obs_params=self.sim_mgr.obs_params,
                best=mean_test_score > best_score,
            )
            best_score = max(best_score, mean_test_score)
            self._logger.info(
                'Training done, best_score={0:.4f}'.format(best_score))

            return best_score
